import 'dart:async';
import 'dart:io';

import 'package:embrace_platform_interface/embrace_platform_interface.dart';
import 'package:embrace_platform_interface/http_method.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';

export 'package:embrace_platform_interface/http_method.dart' show HttpMethod;

/// Entry point for the SDK. This class is part of the Embrace Public API.
class Embrace {
  // ignore: empty_constructor_bodies
  Embrace._() {}
  EmbracePlatform get _platform => EmbracePlatform.instance;
  static final Embrace _instance = Embrace._();

  /// Entry point for the SDK. This class is part of the Embrace Public API.
  static Embrace get instance => _instance;

  /// Starts instrumentation of Dart code by the Embrace SDK. This should be
  /// wrap the entire contents of your Dart main() function if you
  /// wish to capture Dart errors.
  ///
  /// You must also call Embrace.instance.start as soon as possible in your
  /// Android and iOS native code.
  Future<void> start(
    FutureOr<void> Function() action, {
    bool enableIntegrationTesting = false,
  }) {
    return _start(action, enableIntegrationTesting);
  }

  /// Signals that the app has completed startup.
  void endStartupMoment({Map<String, String>? properties}) {
    _runCatching(
      'endStartupMoment',
      () => {_platform.endAppStartup(properties)},
    );
  }

  /// Logs a breadcrumb.
  ///
  /// Breadcrumbs track a user's journey through the application and will be
  /// shown on the timeline.
  void logBreadcrumb(String message) {
    _runCatching(
      'logBreadcrumb',
      () => {_platform.logBreadcrumb(message)},
    );
  }

  /// Remotely logs a message at INFO level
  void logInfo(String message, {Map<String, String>? properties}) {
    _runCatching(
      'logInfo',
      () => {_platform.logInfo(message, properties)},
    );
  }

  /// Remotely logs a message at WARNING level
  void logWarning(
    String message, {
    Map<String, String>? properties,
    bool allowScreenshot = false,
  }) {
    _runCatching(
      'logWarning',
      () => {
        _platform.logWarning(
          message,
          properties,
          allowScreenshot: allowScreenshot,
        )
      },
    );
  }

  /// Remotely logs a message at ERROR level
  void logError(
    String message, {
    Map<String, String>? properties,
    bool allowScreenshot = false,
  }) {
    _runCatching(
      'logError',
      () => {
        _platform.logError(
          message,
          properties,
          allowScreenshot: allowScreenshot,
        )
      },
    );
  }

  /// Log a network request.
  ///
  /// Use the [error] parameter to pass the reason phrase or exception message
  /// for requests that are not successful.
  void logNetworkRequest({
    required String url,
    required HttpMethod method,
    required int startTime,
    required int endTime,
    required int bytesSent,
    required int bytesReceived,
    required int statusCode,
    String? error,
    String? traceId,
  }) {
    _runCatching(
      'logNetworkRequest',
      () => {
        _platform.logNetworkRequest(
          url: url,
          method: method,
          startTime: startTime,
          endTime: endTime,
          bytesSent: bytesSent,
          bytesReceived: bytesReceived,
          statusCode: statusCode,
          error: error,
          traceId: traceId,
        )
      },
    );
  }

  /// Starts a moment.
  ///
  /// Moments are used for encapsulating particular activities within the app,
  /// such as the user adding an item to their shopping cart. The length of time
  /// a moment takes to execute is recorded, and a screenshot can be taken if a
  /// moment is 'late'.
  void startMoment(
    String name, {
    String? identifier,
    bool allowScreenshot = false,
    Map<String, String>? properties,
  }) {
    _runCatching(
      'startMoment',
      () => {
        _platform.startMoment(
          name,
          identifier,
          properties,
          allowScreenshot: allowScreenshot,
        )
      },
    );
  }

  /// Signals the end of a moment with the specified [name] and [identifier].
  ///
  /// The duration of the moment is computed, and a screenshot taken
  /// (if enabled) if the moment was late.
  void endMoment(
    String name, {
    String? identifier,
    Map<String, String>? properties,
  }) {
    _runCatching(
      'endMoment',
      () => {_platform.endMoment(name, identifier, properties)},
    );
  }

  /// Log the start of a view.
  ///
  /// A matching call to [endView] must be made.
  void startView(String name) {
    _runCatching(
      'startView',
      () => {_platform.startView(name)},
    );
  }

  /// Log the end of a view.
  ///
  /// A matching call to [startView] must be made before ending the view.
  void endView(String name) {
    _runCatching(
      'endView',
      () => {_platform.endView(name)},
    );
  }

  /// Sets the user identifier
  ///
  /// This would typically be some form of unique identifier such as a UUID or
  /// database key for the user.
  /// Embrace will generate a per-install UUID by default.
  void setUserIdentifier(String id) {
    _runCatching(
      'setUserIdentifier',
      () => {_platform.setUserIdentifier(id)},
    );
  }

  /// Clears the currently set user identifier.
  void clearUserIdentifier() {
    _runCatching(
      'clearUserIdentifier',
      () => {_platform.clearUserIdentifier()},
    );
  }

  /// Sets the current user's name.
  void setUserName(String name) {
    _runCatching(
      'setUserName',
      () => {_platform.setUserName(name)},
    );
  }

  /// Clear the currently set user name.
  void clearUserName() {
    _runCatching(
      'clearUserName',
      () => {_platform.clearUserName()},
    );
  }

  /// Sets the current user's email.
  void setUserEmail(String email) {
    _runCatching(
      'setUserEmail',
      () => {_platform.setUserEmail(email)},
    );
  }

  /// Clear the currently set user email.
  void clearUserEmail() {
    _runCatching(
      'clearUserEmail',
      () => {_platform.clearUserEmail()},
    );
  }

  /// Sets this user as a paying user.
  ///
  /// This adds a persona to the user's identity.
  void setUserAsPayer() {
    _runCatching(
      'setUserAsPayer',
      () => {_platform.setUserAsPayer()},
    );
  }

  /// Clears this user as a paying user.
  ///
  /// This would typically be called if a user is no longer paying for
  /// the service and has reverted back to a basic user.
  void clearUserAsPayer() {
    _runCatching(
      'clearUserAsPayer',
      () => {_platform.clearUserAsPayer()},
    );
  }

  /// Sets a custom user persona.
  ///
  /// A persona is a trait associated with a given user.
  void setUserPersona(String persona) {
    _runCatching(
      'setUserPersona',
      () => {_platform.setUserPersona(persona)},
    );
  }

  /// Clears the custom user persona, if it is set.
  void clearUserPersona(String persona) {
    _runCatching(
      'clearUserPersona',
      () => {_platform.clearUserPersona(persona)},
    );
  }

  /// Clears all custom user personas from the user.
  void clearAllUserPersonas() {
    _runCatching(
      'clearAllUserPersonas',
      () => {_platform.clearAllUserPersonas()},
    );
  }

  /// Get the Embrace user identifier assigned to the device.
  Future<String?> getDeviceId() {
    return _platform.getDeviceId();
  }

  /// Manually logs a Dart error or exception to Embrace. You should use this
  /// if you want to capture errors/exceptions and report them to Embrace.
  /// A good example would be to call this function from within a try-catch
  /// block.
  ///
  /// It is not necessary to call this function for uncaught Flutter errors
  /// or Dart errors originating from the root isolate - Embrace automatically
  /// captures these errors already.
  void logDartError(Object error, StackTrace stack) {
    EmbracePlatform.instance
        .logDartError(stack.toString(), error.toString(), null, null);
  }
}

/// Runs an action and catches any exception/error. If an exception/error is
/// thrown it will be reported as an internal error to Embrace.
void _runCatching(String message, void Function() action) {
  try {
    action();
  } catch (e) {
    EmbracePlatform.instance.logInternalError(message, e.toString());
  }
}

Future<void> _start(
  FutureOr<void> Function() action,
  bool enableIntegrationTesting,
) async {
  // step 1 - ensure channels are initialized before calling runApp()
  WidgetsFlutterBinding.ensureInitialized();

  // step 2 - attach to the host SDK
  await EmbracePlatform.instance
      .attachToHostSdk(enableIntegrationTesting: enableIntegrationTesting);

  // step 2 - install a Flutter error handler
  _installFlutterOnError();

  // step 3 - attach to the host SDK
  await EmbracePlatform.instance
      .attachToHostSdk(enableIntegrationTesting: enableIntegrationTesting);

  // step 4 - run everything in a Zone & call runApp() supplied by developer
  _installGlobalErrorHandler(action);
}

/// Installs a Flutter.onError handler to capture uncaught Dart errors/
/// exceptions that originate from within Flutter.
void _installFlutterOnError() {
  // unless a user has explicitly set it to null to ignore errors (which is
  // recommended against), onError should never be null.
  final prevHandler = FlutterError.onError;

  FlutterError.onError = (FlutterErrorDetails details) {
    _processFlutterError(details);

    if (prevHandler != null) {
      prevHandler(details);
    } else {
      // fallback behaviour in the unlikely case onError was set to null.
      FlutterError.presentError(details);
      if (kReleaseMode) {
        exit(1);
      }
    }
  };
}

/// Installs a guarded zone to capture any uncaught Dart errors/
/// exceptions that originate from outside of the Flutter callstack.
/// https://api.flutter.dev/flutter/dart-async/runZonedGuarded.html
void _installGlobalErrorHandler(FutureOr<void> Function() action) {
  runZonedGuarded<Future<void>>(
    () async {
      await action();
    },
    _processGlobalZoneError,
  );
}

/// Processes an error caught in Flutter.onError.
void _processFlutterError(FlutterErrorDetails details) {
  EmbracePlatform.instance.logDartError(
    details.stack?.toString(),
    details.summary.toString(),
    details.context?.toString(),
    details.library,
  );
}

/// Processes an error caught in Embrace's global zone.
void _processGlobalZoneError(Object error, StackTrace stack) {
  EmbracePlatform.instance.logDartError(
    stack.toString(),
    error.toString(),
    null,
    null,
  );
}
