import 'package:embrace/embrace.dart';
import 'package:embrace_platform_interface/embrace_platform_interface.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mocktail/mocktail.dart';
import 'package:plugin_platform_interface/plugin_platform_interface.dart';

class MockEmbracePlatform extends Mock
    with MockPlatformInterfaceMixin
    implements EmbracePlatform {}

const errorMessage = '__error__';

class MockError {
  @override
  String toString() => errorMessage;
}

void main() {
  TestWidgetsFlutterBinding.ensureInitialized();

  group('Embrace', () {
    late EmbracePlatform embracePlatform;

    setUp(() {
      embracePlatform = MockEmbracePlatform();
      EmbracePlatform.instance = embracePlatform;
    });

    group('start', () {
      setUp(() {
        when(
          () => embracePlatform.attachToHostSdk(
            enableIntegrationTesting: any(named: 'enableIntegrationTesting'),
          ),
        ).thenAnswer((_) async => true);
      });

      test('calls the supplied action', () async {
        var count = 0;
        await Embrace.instance.start(() => count++);
        expect(count, 1);
      });

      test('does nothing if action throws an error', () {
        Embrace.instance.start(
          () => throw Error(),
        );
        // TODO(fractalwrench): test error handling once implemented.
      });

      test('attaches to the host sdk when platform implementation exists', () {
        const enableIntegrationTesting = true;
        Embrace.instance
            .start(() {}, enableIntegrationTesting: enableIntegrationTesting);
        verify(
          () => embracePlatform.attachToHostSdk(
            enableIntegrationTesting: enableIntegrationTesting,
          ),
        ).called(1);
      });
    });

    group('endStartupMoment', () {
      const properties = {'key': 'value'};
      test('logs an error when platform implementation exists', () {
        Embrace.instance.endStartupMoment(properties: properties);
        verify(() => embracePlatform.endAppStartup(properties)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.endAppStartup(properties))
            .thenThrow(MockError());
        Embrace.instance.endStartupMoment(properties: properties);

        verify(
          () => embracePlatform.logInternalError(
            'endStartupMoment',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('logBreadcrumb', () {
      const message = '__message__';
      test('logs an error when platform implementation exists', () {
        Embrace.instance.logBreadcrumb(message);
        verify(() => embracePlatform.logBreadcrumb(message)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.logBreadcrumb(message))
            .thenThrow(MockError());
        Embrace.instance.logBreadcrumb(message);

        verify(
          () => embracePlatform.logInternalError('logBreadcrumb', errorMessage),
        ).called(1);
      });
    });

    group('logInfo', () {
      const message = '__message__';
      const properties = {'key': 'value'};
      test('logs an error when platform implementation exists', () {
        Embrace.instance.logInfo(message, properties: properties);
        verify(() => embracePlatform.logInfo(message, properties)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.logInfo(message, properties))
            .thenThrow(MockError());
        Embrace.instance.logInfo(message, properties: properties);

        verify(() => embracePlatform.logInternalError('logInfo', errorMessage))
            .called(1);
      });
    });

    group('logWarning', () {
      const message = '__message__';
      const properties = {'key': 'value'};
      const allowScreenshots = true;
      test('logs a warning when platform implementation exists', () {
        Embrace.instance.logWarning(
          message,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );
        verify(
          () => embracePlatform.logWarning(
            message,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.logWarning(
            message,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).thenThrow(MockError());
        Embrace.instance.logWarning(
          message,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );

        verify(
          () => embracePlatform.logInternalError(
            'logWarning',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('logError', () {
      const message = '__message__';
      const properties = {'key': 'value'};
      const allowScreenshots = true;
      test('logs an error when platform implementation exists', () {
        Embrace.instance.logError(
          message,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );
        verify(
          () => embracePlatform.logError(
            message,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.logError(
            message,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).thenThrow(MockError());
        Embrace.instance.logError(
          message,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );

        verify(
          () => embracePlatform.logInternalError(
            'logError',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('logNetworkRequest', () {
      const url = '__url__';
      const method = HttpMethod.get;
      const startTime = 123;
      const endTime = 321;
      const bytesSent = 222;
      const bytesReceived = 333;
      const statusCode = 200;
      const error = '__error__';
      const traceId = '__traceId__';
      test(
          'logs the specified network request when '
          'platform implementation exists', () {
        Embrace.instance.logNetworkRequest(
          url: url,
          method: method,
          startTime: startTime,
          endTime: endTime,
          bytesSent: bytesSent,
          bytesReceived: bytesReceived,
          statusCode: statusCode,
          error: error,
          traceId: traceId,
        );
        verify(
          () => embracePlatform.logNetworkRequest(
            url: url,
            method: method,
            startTime: startTime,
            endTime: endTime,
            bytesSent: bytesSent,
            bytesReceived: bytesReceived,
            statusCode: statusCode,
            error: error,
            traceId: traceId,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.logNetworkRequest(
            url: url,
            method: method,
            startTime: startTime,
            endTime: endTime,
            bytesSent: bytesSent,
            bytesReceived: bytesReceived,
            statusCode: statusCode,
            error: error,
            traceId: traceId,
          ),
        ).thenThrow(MockError());
        Embrace.instance.logNetworkRequest(
          url: url,
          method: method,
          startTime: startTime,
          endTime: endTime,
          bytesSent: bytesSent,
          bytesReceived: bytesReceived,
          statusCode: statusCode,
          error: error,
          traceId: traceId,
        );

        verify(
          () => embracePlatform.logInternalError(
            'logNetworkRequest',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('startMoment', () {
      const momentName = '__momentName__';
      const identifier = '__identifier__';
      const properties = {'key': 'value'};
      const allowScreenshots = true;
      test(
          'starts the specified moment when '
          'platform implementation exists', () {
        Embrace.instance.startMoment(
          momentName,
          identifier: identifier,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );
        verify(
          () => embracePlatform.startMoment(
            momentName,
            identifier,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.startMoment(
            momentName,
            identifier,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).thenThrow(MockError());
        Embrace.instance.startMoment(
          momentName,
          identifier: identifier,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );

        verify(
          () => embracePlatform.logInternalError(
            'startMoment',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('endMoment', () {
      const momentName = '__momentName__';
      const identifier = '__identifier__';
      const properties = {'key': 'value'};
      test('ends the specified moment when platform implementation exists', () {
        Embrace.instance.endMoment(
          momentName,
          identifier: identifier,
          properties: properties,
        );
        verify(
          () => embracePlatform.endMoment(
            momentName,
            identifier,
            properties,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.endMoment(
            momentName,
            identifier,
            properties,
          ),
        ).thenThrow(MockError());
        Embrace.instance.endMoment(
          momentName,
          identifier: identifier,
          properties: properties,
        );

        verify(
          () => embracePlatform.logInternalError(
            'endMoment',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('startView', () {
      const view = '__view__';
      test('starts the specified view when platform implementation exists', () {
        Embrace.instance.startView(view);
        verify(() => embracePlatform.startView(view)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.startView(view)).thenThrow(MockError());
        Embrace.instance.startView(view);

        verify(
          () => embracePlatform.logInternalError(
            'startView',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('endView', () {
      const view = '__view__';
      test('ends the specified view when platform implementation exists', () {
        Embrace.instance.endView(view);
        verify(() => embracePlatform.endView(view)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.endView(view)).thenThrow(MockError());
        Embrace.instance.endView(view);

        verify(
          () => embracePlatform.logInternalError(
            'endView',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserIdentifier', () {
      const userIdentifier = '__userIdentifier__';
      test('sets the user name when platform implementation exists', () {
        Embrace.instance.setUserIdentifier(userIdentifier);
        verify(() => embracePlatform.setUserIdentifier(userIdentifier))
            .called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.setUserIdentifier(userIdentifier))
            .thenThrow(MockError());
        Embrace.instance.setUserIdentifier(userIdentifier);

        verify(
          () => embracePlatform.logInternalError(
            'setUserIdentifier',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserIdentifier', () {
      test(
          'clears the user identifier when '
          'platform implementation exists', () {
        Embrace.instance.clearUserIdentifier();
        verify(embracePlatform.clearUserIdentifier).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearUserIdentifier).thenThrow(MockError());
        Embrace.instance.clearUserIdentifier();

        verify(
          () => embracePlatform.logInternalError(
            'clearUserIdentifier',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserName', () {
      const userName = '__userName__';
      test('sets the user name when platform implementation exists', () {
        Embrace.instance.setUserName(userName);
        verify(() => embracePlatform.setUserName(userName)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.setUserName(userName))
            .thenThrow(MockError());
        Embrace.instance.setUserName(userName);

        verify(
          () => embracePlatform.logInternalError(
            'setUserName',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserName', () {
      test('sets the user email when platform implementation exists', () {
        Embrace.instance.clearUserName();
        verify(embracePlatform.clearUserName).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearUserName).thenThrow(MockError());
        Embrace.instance.clearUserName();

        verify(
          () => embracePlatform.logInternalError(
            'clearUserName',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserEmail', () {
      const email = '__email__';
      test('sets the user email when platform implementation exists', () {
        Embrace.instance.setUserEmail(email);
        verify(() => embracePlatform.setUserEmail(email)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.setUserEmail(email)).thenThrow(MockError());
        Embrace.instance.setUserEmail(email);

        verify(
          () => embracePlatform.logInternalError(
            'setUserEmail',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserEmail', () {
      test('clears the user email when platform implementation exists', () {
        Embrace.instance.clearUserEmail();
        verify(embracePlatform.clearUserEmail).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearUserEmail).thenThrow(MockError());
        Embrace.instance.clearUserEmail();

        verify(
          () => embracePlatform.logInternalError(
            'clearUserEmail',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserAsPayer', () {
      test('sets the user as payer when platform implementation exists', () {
        Embrace.instance.setUserAsPayer();
        verify(embracePlatform.setUserAsPayer).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.setUserAsPayer).thenThrow(MockError());
        Embrace.instance.setUserAsPayer();

        verify(
          () => embracePlatform.logInternalError(
            'setUserAsPayer',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserAsPayer', () {
      test('clears the user as payer when platform implementation exists', () {
        Embrace.instance.clearUserAsPayer();
        verify(embracePlatform.clearUserAsPayer).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearUserAsPayer).thenThrow(MockError());
        Embrace.instance.clearUserAsPayer();

        verify(
          () => embracePlatform.logInternalError(
            'clearUserAsPayer',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserPersona', () {
      const persona = '__persona__';
      test('sets the user persona when platform implementation exists', () {
        Embrace.instance.setUserPersona(persona);
        verify(() => embracePlatform.setUserPersona(persona)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.setUserPersona(persona))
            .thenThrow(MockError());
        Embrace.instance.setUserPersona(persona);

        verify(
          () => embracePlatform.logInternalError(
            'setUserPersona',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserPersona', () {
      const persona = '__persona__';
      test('clears the user persona when platform implementation exists', () {
        Embrace.instance.clearUserPersona(persona);
        verify(() => embracePlatform.clearUserPersona(persona)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.clearUserPersona(persona))
            .thenThrow(MockError());
        Embrace.instance.clearUserPersona(persona);

        verify(
          () => embracePlatform.logInternalError(
            'clearUserPersona',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearAllUserPersonas', () {
      test('clears user personas when platform implementation exists', () {
        Embrace.instance.clearAllUserPersonas();
        verify(embracePlatform.clearAllUserPersonas).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearAllUserPersonas).thenThrow(MockError());
        Embrace.instance.clearAllUserPersonas();

        verify(
          () => embracePlatform.logInternalError(
            'clearAllUserPersonas',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('getDeviceId', () {
      test('returns device ID when platform implementation exists', () async {
        const platformName = '__test_platform__';
        when(embracePlatform.getDeviceId).thenAnswer((_) async => platformName);

        final actualPlatformName = await Embrace.instance.getDeviceId();
        expect(actualPlatformName, equals(platformName));
      });
    });
  });
}
